import nibabel as nib
import numpy as np


def get_tkr_matrix(shape, zooms):
    tkrcosines = np.array([[-1, 0, 0], [0, 0, 1], [0, -1, 0]])
    mat = tkrcosines * zooms
    # noinspection PyUnresolvedReferences
    return nib.affines.from_matvec(mat, -mat @ shape / 2)



from copy import deepcopy
from pathlib import Path
from string import Template
import secrets
from typing import List

import numpy as np
from nibabel.volumeutils import array_to_file
from nibabel.openers import ImageOpener

from .nib_parrec_fork import PARRECImage, PARRECHeader, PARRECArrayProxy, vol_numbers


def array_string_func(format_str, sep=" "):
    return lambda x: sep.join([format_str] * len(x)).format(
        *tuple(x.squeeze().tolist()) if len(x) > 1 else (x[0],)
    )


def numpy_str_replace():
    return lambda x: "{:s}".format((x if type(x) == str else x.decode("UTF-8")).replace(" ", "-"))


def gen_dict_strings(format_dict, value_dict_array):
    if isinstance(value_dict_array, np.record):
        value_dict_array = {k: v for k, v in zip(value_dict_array.dtype.names, value_dict_array)}
    return {
        k: func(value_dict_array.get(k, default))
        if callable(func)
        else func.format(value_dict_array.get(k, default))
        for k, (func, default) in format_dict.items()
    }


gen_info_types = {
    "patient_name": ("{:s}", ""),
    "exam_name": ("{:s}", ""),
    "protocol_name": ("{:s}", ""),
    "exam_date": ("{:s}", ""),
    "series_type": ("{:s}", ""),
    "acq_nr": ("{:d}", 0),
    "recon_nr": ("{:d}", 0),
    "scan_duration": ("{:.1f}", 0.0),
    "max_cardiac_phases": ("{:d}", 0),
    "max_echoes": ("{:d}", 0),
    "max_slices": ("{:d}", 0),
    "max_dynamics": ("{:d}", 0),
    "max_mixes": ("{:d}", 0),
    "patient_position": ("{:s}", ""),
    "prep_direction": ("{:s}", ""),
    "tech": ("{:s}", ""),
    "scan_resolution": (array_string_func("{:d}"), np.array([0.0, 0.0])),
    "scan_mode": ("{:s}", ""),
    "repetition_time": (array_string_func("{:.3f}"), np.array([0.0])),
    "fov": (array_string_func("{:.3f}"), np.array([0.0, 0.0, 0.0])),
    "water_fat_shift": ("{:.3f}", 0.0),
    "angulation": (array_string_func("{:.3f}"), np.array([0.0, 0.0, 0.0])),
    "off_center": (array_string_func("{:.3f}"), np.array([0.0, 0.0, 0.0])),
    "flow_compensation": ("{:d}", 0),
    "presaturation": ("{:d}", 0),
    "phase_enc_velocity": (array_string_func("{:.6f}"), np.array([0.0, 0.0, 0.0])),
    "mtc": ("{:d}", 0),
    "spir": ("{:d}", 0),
    "epi_factor": ("{:d}", 1),
    "dyn_scan": ("{:d}", 0),
    "diffusion": ("{:d}", 0),
    "diffusion_echo_time": ("{:.4f}", 0.0),
    "max_diffusion_values": ("{:d}", 1),
    "max_gradient_orient": ("{:d}", 1),
    "nr_label_types": ("{:d}", 0),
}

image_def_types = {
    "slice_number": ("{:d}", 0),
    "echo_number": ("{:d}", 0),
    "dynamic_scan_number": ("{:d}", 0),
    "cardiac_phase_number": ("{:d}", 0),
    "image_type_mr": ("{:d}", 0),
    "scanning_sequence": ("{:d}", 0),
    "index_in_rec_file": ("{:d}", 0),
    "image_pixel_size": ("{:d}", 0),
    "scan_percentage": ("{:d}", 0),
    "recon_resolution": (array_string_func("{:d}"), np.array([0, 0])),
    "rescale_intercept": ("{:.5f}", 0.0),
    "rescale_slope": ("{:.5f}", 0.0),
    "scale_slope": ("{:.5e}", 0.0),
    "window_center": (lambda x: "{:d}".format(int(x)), 0.0),
    "window_width": (lambda x: "{:d}".format(int(x)), 0.0),
    "image_angulation": (array_string_func("{:.2f}"), np.array([0.0, 0.0, 0.0])),
    "image_offcentre": (array_string_func("{:.2f}"), np.array([0.0, 0.0, 0.0])),
    "slice_thickness": ("{:.3f}", 0.0),
    "slice_gap": ("{:.3f}", 0.0),
    "image_display_orientation": ("{:d}", 0),
    "slice_orientation": ("{:d}", 0),
    "fmri_status_indication": ("{:d}", 0),
    "image_type_ed_es": ("{:d}", 0),
    "pixel_spacing": (array_string_func("{:.3f}"), np.array([0.0, 0.0])),
    "echo_time": ("{:.2f}", 0.0),
    "dyn_scan_begin_time": ("{:.2f}", 0.0),
    "trigger_time": ("{:.2f}", 0.0),
    "diffusion_b_factor": ("{:.2f}", 0.0),
    "number_of_averages": ("{:d}", 0),
    "image_flip_angle": ("{:.2f}", 0.0),
    "cardiac_frequency": ("{:d}", 0),
    "minimum_rr_interval": ("{:d}", 0),
    "maximum_rr_interval": ("{:d}", 0),
    "turbo_factor": ("{:d}", 0),
    "inversion_delay": ("{:.1f}", 0.0),
    "diffusion_b_value_number": ("{:d}", 0),
    "gradient_orientation_number": ("{:d}", 0),
    "contrast_type": ("{:d}", 0),
    "diffusion_anisotropy_type": ("{:d}", 0),
    "diffusion": (array_string_func("{:.3f}"), np.array([0.0, 0.0, 0.0])),
    "label_type": ("{:d}", 0),
    "contrast_bolus_agent": (numpy_str_replace(), ""),
    "contrast_bolus_route": (numpy_str_replace(), ""),
    "contrast_bolus_volume": ("{:.6f}", 0.0),
    "contrast_bolus_start_time": (numpy_str_replace(), ""),
    "contrast_bolus_total_dose": ("{:.6f}", 0.0),
    "contrast_bolus_ingredient": (numpy_str_replace(), ""),
    "contrast_bolus_ingredient_concentration": ("{:.6f}", 0.0),
}

parrec_templates = Path(__file__).parent / "parrec_templates"

top_header_template = Template((parrec_templates / "top_header.txt").read_text())
gen_info_template = Template((parrec_templates / "gen_info.txt").read_text())
image_def_template = Template("  ".join(["$" + k for k in image_def_types]) + "\n")


def generate_par_file(dataset_name: str, header: PARRECHeader, filename: Path) -> None:
    with filename.open("w") as fp:
        fp.write(top_header_template.substitute({"dataset_name": dataset_name}))
        fp.write(
            gen_info_template.substitute(gen_dict_strings(gen_info_types, header.general_info))
        )
        fp.write((parrec_templates / "pixel_values.txt").read_text())
        image_defs = header.image_defs.view(np.recarray)
        for i in range(len(image_defs)):
            fp.write(
                image_def_template.substitute(gen_dict_strings(image_def_types, image_defs[i]))
            )
        fp.write(
            "\n# === END OF DATA DESCRIPTION FILE ===============================================\n"
        )


def split_fix_parrec(in_filename: Path, study_uid, outdir) -> List[str]:
    file_map = PARRECImage.filespec_to_file_map(str(in_filename))
    with file_map["header"].get_prepare_fileobj("rt") as hdr_fobj:
        hdr = PARRECHeader.from_fileobj(hdr_fobj, permit_truncated=False, strict_sort=False)

    idefs = hdr.image_defs
    split_types = ["echo_number", "image_type_mr"]
    slice_vols = np.array(vol_numbers(idefs["slice_number"]))
    split_defs = set(zip(zip(*[idefs[def_name].tolist() for def_name in split_types]), slice_vols))
    split_vols = {}
    for split_def, vol_num in split_defs:
        if split_def not in split_vols:
            split_vols[split_def] = []
        split_vols[split_def].append(vol_num)

    series_uid = (
        study_uid
        + (".%02d.%02d." % (hdr.general_info["acq_nr"], hdr.general_info["recon_nr"]))
        + str(int(str(secrets.randbits(16))))
    )
    for i, vol_nums in enumerate(sorted(split_vols.values())):
        this_hdr = deepcopy(hdr)
        this_hdr.image_defs = idefs[np.in1d(slice_vols, vol_nums)]
        this_hdr.image_defs["echo_number"] = [1] * len(this_hdr.image_defs["echo_number"])
        generate_par_file(
            str(in_filename), this_hdr, outdir / (series_uid + (".%02d" % (i + 1)) + ".par")
        )
    if len(split_vols) == 1:
        Path(file_map["image"].filename).rename(outdir / (series_uid + ".01.rec"))
    else:
        rec_fobj = file_map["image"].get_prepare_fileobj()
        data = PARRECArrayProxy(rec_fobj, hdr).get_unscaled()
        for i, vol_nums in enumerate(sorted(split_vols.values())):
            with ImageOpener(
                outdir / (series_uid + (".%02d" % (i + 1)) + ".rec"), mode="wb"
            ) as fileobj:
                array_to_file(
                    data[..., np.in1d(range(data.shape[-1]), vol_nums)],
                    fileobj,
                    hdr.get_data_dtype(),
                    order="F",
                )

    return [series_uid + (".%02d" % (i + 1)) + ".par" for i in range(len(split_vols))] + [
        series_uid + (".%02d" % (i + 1)) + ".rec" for i in range(len(split_vols))
    ]



from __future__ import annotations
from collections import defaultdict
import datetime
import json
import logging
from pathlib import Path
import re
import secrets
import shutil
from subprocess import run, PIPE, STDOUT
from typing import Callable, List, Tuple, Union, Any, Optional

import numpy as np

from .._version import __version__
from ..qa.create import create_qa_image
from .json import NoIndent, JSONObjectEncoder
from .lut import LookupTable
from .metadata import Metadata

# from .logging import WARNING_DEBUG
from .utils import (
    mkdir_p,
    reorient,
    parse_dcm2niix_filenames,
    remove_created_files,
    hash_file_list,
    none_to_float,
    find_closest,
    FILE_OCTAL,
    hash_file_dir,
    p_add,
    get_software_versions,
    hash_value,
    shift_date,
    parse_dcm2niix_suffixes,
)

DESCRIPTION_IGNORE = ["loc", "survey", "scout", "3-pl", "scanogram", "smartbrain", "pride"]
POSTGAD_DESC = ["post", "+c", "gad", "gd", "pstc", "+ c", "c+"]
MATCHING_ITEMS = [
    "ImageOrientationPatient",
    "RepetitionTime",
    "FlipAngle",
    "EchoTime",
    "TriggerTime",
    "InversionTime",
    "ComplexImageComponent",
    "ImageType",
]
HASH_ITEMS = ["InstitutionName", "DeviceIdentifier"]
SHIFT_ITEMS = ["AcqDateTime"]
DCM_ORIENT_PLANES = {0: "sagittal", 1: "coronal", 2: "axial"}
PARREC_ORIENTATIONS = {1: "axial", 2: "sagittal", 3: "coronal"}


class BaseInfo:
    # TODO: Type consistent defaults? Type checking?
    # This can be done using type hints and defaults in the class definition
    def __init__(self, path: Path) -> None:
        self.SourcePath = Path(path.parent.name) / path.name
        self.SourceHash = (
            hash_file_list([path, path.with_suffix(".rec")], include_names=False)
            if self.SourcePath.suffix == ".par"
            else hash_file_dir(path, include_names=False)
        )
        self.SeriesUID = None
        self.StudyUID = None
        self.NumFiles = None
        self.MultiFrame = None
        self.InstitutionName = None
        self.Manufacturer = None
        self.ScannerModelName = None
        self.DeviceIdentifier = None
        self.SeriesDescription = None
        self.AcqDateTime = None
        self.MagneticFieldStrength = None
        self.AcquisitionDimension = None
        self.SliceSpacing = None
        self.SliceThickness = None
        self.FlipAngle = None
        self.RepetitionTime = None
        self.EchoTime = None
        self.InversionTime = None
        self.TriggerTime = None
        self.EchoTrainLength = None
        self.EPIFactor = None
        self.AcquisitionMatrix = []
        self.AcquiredResolution = None
        self.ReconMatrix = None
        self.ReconResolution = None
        self.FieldOfView = None
        self.SequenceType = tuple()
        self.ImageType = None
        self.SeriesNumber = None
        self.ComplexImageComponent = None
        self.BodyPartExamined = None
        self.StudyDescription = None
        self.SequenceVariant = tuple()
        self.ScanOptions = tuple()
        self.SequenceName = None
        self.ExContrastAgent = None
        self.ImageOrientationPatient = None
        self.ImagePositionPatient = None
        self.SliceOrientation = None
        self.SoftwareVersions = None
        self.NumberOfAverages = None
        self.PercentSampling = None
        self.ReceiveCoilName = None
        self.PixelBandwidth = None
        self.VariableFlipAngle = None

        self.ConvertImage = False
        self.NiftiCreated = False
        self.LookupName = None
        self.PredictedName = None
        self.ManualName = None
        self.NiftiName = None
        self.NiftiHash = None

    def __repr_json__(self) -> dict:
        return {k: NoIndent(v) for k, v in self.__dict__.items()}

    def anonymize(self, date_shift_days: int = 0):
        for key in HASH_ITEMS:
            setattr(self, key, hash_value(getattr(self, key)))
        for key in SHIFT_ITEMS:
            setattr(self, key, shift_date(getattr(self, key), date_shift_days))

    def update_name(
        self, name_lambda: Callable[[str], str], message: str = "Adjusting name"
    ) -> None:
        logging.debug(
            "%s for %s: %s --> %s"
            % (message, self.SeriesUID, self.NiftiName, name_lambda(self.NiftiName))
        )
        self.NiftiName = name_lambda(self.NiftiName)

    def should_convert(self) -> bool:
        type_str = " ".join(self.ImageType[:2]).lower()
        series_desc = self.SeriesDescription.lower()
        type_status = ("derived" not in type_str) or (
            "derived" in type_str and "primary" in type_str
        )
        desc_ignore = any([item in series_desc for item in DESCRIPTION_IGNORE]) or re.search(
            r"(?<!cervi)cal(?:\W|ibration|$)", series_desc
        )
        mpr_ignore = (
            (
                (re.search(r".*mpr(?!age).*", series_desc) is not None)
                and self.ImageType[0].lower() != "original"
            )
            or any([img_type.lower() == "mpr" for img_type in self.ImageType])
            or any(["projection" in img_type.lower() for img_type in self.ImageType])
            or "composed" in series_desc
            or any(["composed" in img_type.lower() for img_type in self.ImageType])
        )
        mip_ignore = (
            "mip" in series_desc
            or any([img_type.lower() == "mnip" for img_type in self.ImageType])
            or any([img_type.lower() == "maximum" for img_type in self.ImageType])
        )
        processed_ignore = (
            any(["adc" in img_type.lower() for img_type in self.ImageType])
            or any([img_type.lower() == "sub" for img_type in self.ImageType])
            or any([img_type.lower() == "sum" for img_type in self.ImageType])
            or any([re.search(r"t2.?map", img_type.lower()) for img_type in self.ImageType])
        )
        logging.debug(
            "Derived:%s, Description:%s, MPR:%s, MIP:%s, Processed:%s"
            % (not type_status, desc_ignore, mpr_ignore, mip_ignore, processed_ignore)
        )
        self.ConvertImage = (
            type_status
            and not desc_ignore
            and not mpr_ignore
            and not mip_ignore
            and not processed_ignore
        )
        if not self.ConvertImage:
            logging.info("This series is localizer, derived or processed image. Skipping.")
        return self.ConvertImage

    def automatic_name_generation(self):
        series_desc = self.SeriesDescription.lower()
        if series_desc.startswith("wip"):
            series_desc = series_desc[3:].lstrip()
        scan_opts = [opt.lower() for opt in self.ScanOptions]
        # 1) Orientation
        orientation = self.SliceOrientation.upper() if self.SliceOrientation is not None else "NONE"
        # 2) Resolution
        resolution = self.AcquisitionDimension
        if resolution not in ["2D", "3D"]:
            resolution = "3D" if "3d" in series_desc else "2D"
        # 3) Ex-contrast
        excontrast = "PRE"
        if not (self.ExContrastAgent is None or self.ExContrastAgent.upper() in ["", "NONE"]):
            excontrast = "POST"
        elif any([item in series_desc for item in POSTGAD_DESC]) and "pre" not in series_desc:
            excontrast = "POST"
        # 4) Modality
        desc_modalities = []
        if "t1" in series_desc:
            desc_modalities.append("T1")
        if "t2" in series_desc:
            desc_modalities.append("T2")
        if "flair" in series_desc:
            desc_modalities.append("FLAIR")
        if (
            re.search(r"medic(?!al)", series_desc)
            or "t2star" in series_desc
            or re.search(r"swi(?!p)", series_desc)
            or "t2*" in series_desc
            or "swan" in series_desc
        ):
            desc_modalities.append("T2STAR")
        if "stir" in series_desc:
            desc_modalities.append("STIR")
        if (
            "dti" in series_desc
            or "diff" in series_desc
            or re.search(r"(?<!p)dw", series_desc)
            or series_desc.endswith("b1000")
            or series_desc.endswith("b0")
            or "tensor" in series_desc
            or any([img_type.lower() == "diffusion" for img_type in self.ImageType])
        ):
            desc_modalities.append("DIFF")

        modality = "UNK"
        if len(desc_modalities) == 1:
            modality = desc_modalities[0]
        else:
            if "T1" in desc_modalities and "T2STAR" in desc_modalities:
                modality = "T1"
            if "T1" in desc_modalities and "FLAIR" in desc_modalities:
                modality = "T1"
            if "T2" in desc_modalities and "T2STAR" in desc_modalities:
                modality = "T2STAR"
        if series_desc in ["swi_images", "pha_images", "mag_images"]:
            self.ComplexImageComponent = {
                "swi_images": "SWI",
                "mag_images": "MAGNITUDE",
                "pha_images": "PHASE",
            }[series_desc]
            modality = "T2STAR"
        elif any(["flow" in img_type.lower() for img_type in self.ImageType]) or any(
            ["velocity" in img_type.lower() for img_type in self.ImageType]
        ):
            modality = "FLOW"
        elif (
            any(["tof" in img_type.lower() for img_type in self.ImageType])
            or any(["tof" in variant.lower() for variant in self.SequenceVariant])
            or re.search(r"tof(?!f)", series_desc)
            or "angio" in series_desc
        ):
            modality = "TOF"
        elif (
            any(["mtc" in variant.lower() for variant in self.SequenceVariant])
            or getattr(self, "MTContrast", 0) == 1
            or "mt_gems" in scan_opts
        ):
            modality = "MT"
        elif getattr(self, "DiffusionFlag", 0) == 1:
            modality = "DIFF"
        # 5) Sequence
        etl = 1 if self.EchoTrainLength is None else self.EchoTrainLength
        seq_name = "" if self.SequenceName is None else self.SequenceName.lower()
        seq_type = [seq.lower() for seq in self.SequenceType]
        seq_var = [variant.lower() for variant in self.SequenceVariant]
        sequence = "UNK"
        if any([seq == "se" for seq in seq_type]):
            sequence = "SE"
        elif any([seq == "gr" for seq in seq_type]):
            sequence = "GRE"
        elif self.FlipAngle >= 60:
            sequence = "SE"
        if (
            any(["ep" == seq for seq in seq_type])
            or "epi" in seq_name
            or (self.EPIFactor is not None and self.EPIFactor > 1)
        ):
            sequence = "GRASE" if seq_name == "grase" else "EPI"
        if re.search(r"(t1.?[tf]fe|fl3d1)", seq_name):
            sequence = "SPGR"
        if sequence == "GRE" and (
            any([variant == "sp" for variant in seq_var])
            or any([variant == "ss" for variant in seq_var])
        ):
            sequence = "SPGR"
        if sequence not in ["EPI", "GRASE"] and (
            etl > 1 or "fast_gems" in scan_opts or "fse" in seq_name
        ):
            sequence = "F" + sequence
        if sequence not in ["EPI", "GRASE"] and "IR" not in sequence:
            if (
                (self.InversionTime is not None and self.InversionTime > 50)
                or any([seq == "ir" for seq in seq_type])
                or any([variant == "mp" for variant in seq_var])
                or re.search(r"t1.?tfe", seq_name)
                or re.search(r"(flair|stir|mp.?rage|bravo)", series_desc)
            ):
                sequence = "IR" + sequence
        if sequence.startswith("IR") and resolution == "3D" and "F" not in sequence:
            sequence = sequence.replace("IR", "IRF")
        if sequence == "IRFGRE" or (sequence == "IRFUNK" and modality == "T1"):
            sequence = "IRFSPGR"
        if modality == "UNK":
            if sequence.endswith("SE"):
                modality = "T2"
            elif sequence.endswith("GRE") or sequence.endswith("SPGR"):
                modality = "T1" if self.EchoTime is not None and self.EchoTime < 15 else "T2STAR"
        if modality == "T2" and sequence.startswith("IR"):
            modality = (
                "STIR" if self.InversionTime is not None and self.InversionTime < 400 else "FLAIR"
            )
        elif modality == "T2" and (sequence.endswith("GRE") or sequence.endswith("SPGR")):
            modality = "T2STAR"
        elif modality == "T2" and self.EchoTime is not None and self.EchoTime < 30:
            modality = (
                "PD" if self.RepetitionTime is not None and self.RepetitionTime > 800 else "T1"
            )

        body_part = "BRAIN"
        body_part_ex = "" if self.BodyPartExamined is None else self.BodyPartExamined.lower()
        study_desc = (
            "" if self.StudyDescription is None else self.StudyDescription.lower().replace(" ", "")
        )
        # Define a list of tuples with regular expression patterns and corresponding body part values
        patterns = [
            (r"(brain|^br_)", "BRAIN"),
            (r"ct[ -]?spine", "SPINE"),
            (
                r"(?<!\w)(cerv|c[ -]?sp|c.?spine|msma)",
                "CSPINE",
                r"(?<!\w)(thor|t[ -]?sp|t.?spine)",
                "SPINE",
            ),
            (
                r"(?<!\w)(thor|t[ -]?sp|t.?spine)",
                "TSPINE",
                r"(?<!\w)(cerv|(?<!ci)c[ -]?sp|(?<!ci)c.?spine)",
                "SPINE",
            ),
            (r"(?<!\w)(lumb|l[ -]?sp|l.?spine)", "LSPINE"),
            (r"(\sc.?tl?(?:\s+|$)|^sp_|t1.ax.vibe|t1.vibe.tra|ax.t1.vibe)", "SPINE"),
            (r"(orbit|thin|^on_)", "ORBITS"),
            (r"spine", "SPINE"),
        ]

        # Iterate through patterns and match against relevant variables
        found = False
        for i, search_str in enumerate([series_desc, body_part_ex, study_desc]):
            for pattern in patterns:
                if re.search(pattern[0], search_str):
                    body_part = pattern[1]
                    if i > 0 and len(pattern) > 2 and re.search(pattern[2], study_desc):
                        body_part = pattern[3]
                    found = True
                    break
            if found:
                break

        if re.search(r"\*?(me2d1r)", seq_name):
            body_part = "SPINE"
        if modality == "DIFF" and orientation == "SAGITTAL":
            body_part = "SPINE"

        slice_sp = (
            float(self.SliceThickness) if self.SliceSpacing is None else float(self.SliceSpacing)
        )
        if (
            self.NumFiles < 10
            and body_part == "BRAIN"
            and modality in ["T1", "T2", "T2STAR", "FLAIR"]
        ):
            logging.info("This series is localizer, derived or processed image. Skipping.")
            return False
        elif modality == "DIFF" and (
            series_desc.endswith("_tracew") or series_desc.endswith("_fa")
        ):
            logging.info("This series is localizer, derived or processed image. Skipping.")
            return False
        elif body_part == "ORBITS" and self.NumFiles * slice_sp > 120:
            body_part = "BRAIN"
        elif body_part == "BRAIN" and self.NumFiles * slice_sp < 100 and orientation == "SAGITTAL":
            body_part = "SPINE"

        if body_part == "SPINE" and re.search(r"upper(?!\s*t)", series_desc):
            body_part = "CSPINE"
        elif body_part == "SPINE" and "lower" in series_desc:
            body_part = "TSPINE"

        return [body_part, modality, sequence, resolution, orientation, excontrast]

    def create_image_name(
        self,
        scan_str: str,
        study_num: int,
        series_num: int,
        lut_obj: LookupTable,
        manual_dict: dict,
    ) -> None:
        source_name = str(self.SourcePath)
        man_list = [None] * 6
        pred_list = [None] * 6
        if source_name in manual_dict:
            man_list = manual_dict[source_name]
        lut_list = lut_obj.check(self.InstitutionName, self.SeriesDescription)
        if man_list is False or lut_list is False:
            self.ConvertImage = False
            return
        if any([item is None for item in man_list]) and any([item is None for item in lut_list]):
            # Needs automatic naming
            logging.debug("Name lookup failed or incomplete, using automatic name generation.")
            try:
                pred_list = self.automatic_name_generation()
            except TypeError:
                logging.exception("Automatic name generation failed.")
                self.ConvertImage = False
                return
        if pred_list is False:
            self.ConvertImage = False
            return

        for item_list in [pred_list, man_list, lut_list]:
            if item_list[1] == "ME":
                if item_list[2].endswith("SE"):
                    item_list[1] = "PD" if self.EchoTime < 30 else "T2"
                else:
                    item_list[1] = "T1" if self.EchoTime < 15 else "T2STAR"

        self.ManualName = man_list
        self.LookupName = lut_list
        self.PredictedName = pred_list

        final_list = [None] * max([len(item_list) for item_list in [man_list, lut_list, pred_list]])
        final_list = [
            self.ManualName[i]
            if i < len(self.ManualName) and final_list[i] is None
            else final_list[i]
            for i in range(len(final_list))
        ]
        final_list = [
            self.LookupName[i]
            if i < len(self.LookupName) and final_list[i] is None
            else final_list[i]
            for i in range(len(final_list))
        ]
        final_list = [
            self.PredictedName[i]
            if i < len(self.PredictedName) and final_list[i] is None
            else final_list[i]
            for i in range(len(final_list))
        ]
        final_list = [item.upper() if item is not None else "UNK" for item in final_list]
        self.NiftiName = "_".join(
            [scan_str, "%02d-%02d" % (study_num, series_num), "-".join(final_list)]
        )
        logging.debug("Predicted name: %s" % self.NiftiName)


class BaseSet:
    def __init__(
        self,
        source: Path,
        output_root: Path,
        metadata_obj: Metadata,
        lut_obj: LookupTable,
        remove_identifiers: bool = False,
        date_shift_days: int = 0,
        manual_names: Optional[dict] = None,
        input_hash: Optional[str] = None,
    ) -> None:
        self.__version__ = {"radifox": __version__, "dcm2niix": get_software_versions()["dcm2niix"]}
        self.ConversionSoftwareVersions = get_software_versions()
        if input_hash is None:
            logging.info("Hashing source file(s) for record keeping.")
            self.InputHash = hash_file_dir(source, False)
            logging.info("Hashing complete.")
        else:
            logging.info("Using existing source hash for record keeping.")
            self.InputHash = input_hash
        self.ManualNames = manual_names if manual_names is not None else {}
        self.Metadata = metadata_obj
        self.LookupTable = lut_obj
        self.RemoveIdentifiers = remove_identifiers
        self.DateShiftDays = date_shift_days
        self.OutputRoot = output_root
        self.SeriesList = []

    def __repr_json__(self) -> dict:
        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in ["OutputRoot", "DateShiftDays"]
        }

    @staticmethod
    def get_unique_study_series(series_list):
        study_nums = {}
        series_nums = {}
        sorted_list = sorted(
            series_list,
            key=lambda x: (
                x.AcqDateTime,
                x.SeriesNumber,
                x.InstitutionName,
                none_to_float(x.MagneticFieldStrength),
                x.ScannerModelName,
            ),
        )
        breaks = []
        for i in range(1, len(sorted_list)):
            if any(
                [
                    getattr(sorted_list[i], key) is None or getattr(sorted_list[i - 1], key) is None
                    for key in ["InstitutionName", "MagneticFieldStrength", "ScannerModelName"]
                ]
            ):
                continue
            if any(
                [
                    getattr(sorted_list[i], key) != getattr(sorted_list[i - 1], key)
                    for key in ["InstitutionName", "MagneticFieldStrength", "ScannerModelName"]
                ]
            ):
                breaks.append(i)
                continue
            t_diff = datetime.datetime.strptime(
                sorted_list[i].AcqDateTime, "%Y-%m-%d %H:%M:%S"
            ) - datetime.datetime.strptime(sorted_list[i - 1].AcqDateTime, "%Y-%m-%d %H:%M:%S")
            if t_diff.total_seconds() > 1800:
                breaks.append(i)
        breaks = [0] + breaks + [len(sorted_list)]
        break_nums = sum([[i] * (breaks[i] - breaks[i - 1]) for i in range(1, len(breaks))], [])
        study_tuples = [(break_nums[i], di.StudyUID) for i, di in enumerate(sorted_list)]
        for i, tup in enumerate(dict((tup, None) for tup in study_tuples).keys()):
            sub_list = [di for j, di in enumerate(sorted_list) if study_tuples[j] == tup]
            sub_series = list(
                dict(((di.SeriesNumber, di.SeriesDescription), None) for di in sub_list).keys()
            )
            for di in sub_list:
                study_nums[di.SourcePath] = i + 1
                series_nums[di.SourcePath] = (
                    sub_series.index((di.SeriesNumber, di.SeriesDescription)) + 1
                )
        return study_nums, series_nums

    def generate_unique_names(self) -> None:
        self.SeriesList = sorted(
            sorted(self.SeriesList, key=lambda x: (x.StudyUID, x.SeriesNumber, x.SeriesUID)),
            key=lambda x: x.ConvertImage,
            reverse=True,
        )
        for i, di in enumerate(self.SeriesList):
            if di.NiftiName is None:
                continue
            # sWIP is Philips indicator for a "sum" of a multi-echo image
            if di.SeriesDescription.startswith("sWIP") or di.SeriesDescription.startswith("smFFE"):
                di.update_name(lambda x: x + "-SUM")
            # ND images for Siemens scans indicates a "no distortion correction" scan
            if (
                di.Manufacturer == "SIEMENS"
                and any([img_type.lower() == "nd" for img_type in di.ImageType])
                and di.SeriesDescription.lower().endswith("_nd")
            ):
                di.update_name(lambda x: x + "-ND")

        # Change T1 image to MT/MTOFF if matches an MT sequence and add MTON for the corresponding MT scan
        for i, di in enumerate(self.SeriesList):
            if di.NiftiName is None:
                continue
            if di.NiftiName.split("_")[-1].split("-")[1] in ["T1", "T2STAR"] and any(
                [
                    di.NiftiName.split("_")[-1]
                    == other_di.NiftiName.split("_")[-1].replace(
                        "-MT-", "-%s-" % di.NiftiName.split("_")[-1].split("-")[1]
                    )
                    for other_di in self.SeriesList
                    if other_di.NiftiName is not None and "-MT-" in other_di.NiftiName
                ]
            ):
                closest_mt = find_closest(
                    i,
                    [
                        j
                        for j, other_di in enumerate(self.SeriesList)
                        if other_di.NiftiName is not None
                        and "-MT-" in other_di.NiftiName
                        and di.NiftiName
                        == other_di.NiftiName.replace(
                            "-MT-", "-%s-" % di.NiftiName.split("_")[-1].split("-")[1]
                        )
                        and di.EchoTime == other_di.EchoTime
                        and di.FlipAngle == other_di.FlipAngle
                        and di.RepetitionTime == other_di.RepetitionTime
                    ],
                )
                if closest_mt is not None:
                    di.update_name(lambda x: x.replace("-T1-", "-MT-") + "-MTOFF")
                    self.SeriesList[closest_mt].update_name(lambda x: x + "-MTON")

        # Change generic spine into CSPINE/TSPINE/LSPINE based on previous image
        spine_indexes = ["SPINE", "CSPINE", "TSPINE", "LSPINE"]
        for study_uid in set(
            [
                di.StudyUID
                for di in self.SeriesList
                if di.NiftiName is not None
                and "SPINE" in di.NiftiName.split("_")[-1].split("-")[0]
                and di.ManualName[0] is None
            ]
        ):
            for series_description in set(
                [
                    di.SeriesDescription
                    for di in self.SeriesList
                    if di.NiftiName is not None
                    and di.StudyUID == study_uid
                    and "SPINE" in di.NiftiName.split("_")[-1].split("-")[0]
                    and di.ManualName[0] is None
                ]
            ):
                di_list = sorted(
                    [
                        di
                        for di in self.SeriesList
                        if di.NiftiName is not None
                        and di.StudyUID == study_uid
                        and di.SeriesDescription == series_description
                        and "SPINE" in di.NiftiName.split("_")[-1].split("-")[0]
                        and di.ManualName[0] is None
                    ],
                    key=lambda x: x.ImagePositionPatient[2],
                    reverse=True,
                )
                spine_idx = 0
                for i, di in enumerate(di_list):
                    current_level = di.NiftiName.split("_")[-1].split("-")[0]
                    if i == 0:
                        if current_level == "SPINE":
                            di.update_name(lambda x: x.replace("_SPINE-", "_CSPINE-"))
                        spine_idx = spine_indexes.index(di.NiftiName.split("_")[-1].split("-")[0])
                        continue
                    if (
                        abs(di.ImagePositionPatient[2] - di_list[i - 1].ImagePositionPatient[2])
                        > 100
                    ):
                        spine_idx = min(spine_idx + 1, len(spine_indexes) - 1)
                    di.update_name(
                        lambda x: x.replace(
                            "_%s-" % current_level, "_%s-" % spine_indexes[spine_idx]
                        )
                    )

        ruid_set = set([".".join(di.SeriesUID.split(".")[:-1]) for di in self.SeriesList])
        ruid_dict = {ruid: [] for ruid in ruid_set}
        for di in self.SeriesList:
            if di.NiftiName is None:
                continue
            root_uid = ".".join(di.SeriesUID.split(".")[:-1])
            ruid_dict[root_uid].append(di)
        dyn_checks = {}
        for root_uid in ruid_dict:
            if len(ruid_dict[root_uid]) > 1:
                dyn_checks[root_uid] = []
                for di in ruid_dict[root_uid]:
                    dyn_checks[root_uid].append(di)
                    dyn_num = dyn_checks[root_uid].index(di) + 1
                    di.update_name(
                        lambda x: x + ("-DYN%d" % dyn_num), "Adding temporary DYN naming"
                    )

        for di_list in dyn_checks.values():
            non_matching = set()
            for item in set(MATCHING_ITEMS) - {"ImageType"}:
                if any(
                    [
                        getattr(di_list[0], item) != getattr(di_list[i], item)
                        for i in range(len(di_list))
                    ]
                ):
                    non_matching.add(item)
            if not non_matching:
                continue

            for di in di_list:
                di.update_name(
                    lambda x: "-".join(x.split("-")[:-1]), "Removing temporary DYN naming"
                )

            if len(set(di.NiftiName.split("_")[-1].split("-")[0] for di in di_list)) == len(
                di_list
            ):
                continue

            if "EchoTime" in non_matching and any(["-T2STAR-" in di.NiftiName for di in di_list]):
                for di in [di for di in di_list if "-T1-" in di.NiftiName]:
                    di.update_name(lambda x: x.replace("-T1-", "-T2STAR-"))

            if "ImageOrientationPatient" in non_matching:
                for i, val in enumerate(
                    sorted(
                        set(di.ImagePositionPatient for di in di_list),
                        key=lambda x: x[2],
                        reverse=True,
                    )
                ):
                    for di in [di for di in di_list if di.ImagePositionPatient == val]:
                        di.update_name(lambda x: x + ("-POS%d" % (i + 1)))
                    non_matching -= {"ImageOrientationPatient"}

            for attr_str, name_str in [("InversionTime", "INV"), ("EchoTime", "ECHO")]:
                if attr_str in non_matching:
                    for i, val in enumerate(sorted(set(getattr(di, attr_str) for di in di_list))):
                        for di in [di for di in di_list if getattr(di, attr_str) == val]:
                            di.update_name(lambda x: x + ("-%s%d" % (name_str, i + 1)))
                    non_matching -= {attr_str}

            if "ComplexImageComponent" in non_matching:
                for di in di_list:
                    di.update_name(lambda x: x + "-" + di.ComplexImageComponent[:3])
                non_matching -= {"ComplexImageComponent"}

            if non_matching:
                for i, di in enumerate(di_list):
                    di.update_name(lambda x: x + ("-DYN%d" % i), "Re-adding extra DYN naming")

        for di in self.SeriesList:
            if di.NiftiName is None:
                continue
            if di.NiftiName.split("_")[-1].split("-")[1] == "T2STAR" and not any(
                [
                    di.NiftiName.split("_")[-1].split("-")[-1] == item
                    for item in ["MAG", "PHA", "SWI"]
                ]
            ):
                comp = (
                    di.ComplexImageComponent[:3] if di.ComplexImageComponent is not None else "MAG"
                )
                di.update_name(lambda x: x + "-" + comp)

        for di in self.SeriesList:
            if di.NiftiName is not None and " ".join(di.ImageType[:2]).lower() == "derived primary":
                if (
                    len(
                        [
                            di2.NiftiName
                            for di2 in self.SeriesList
                            if di2.NiftiName == di.NiftiName
                            and " ".join(di2.ImageType[:2]).lower() == "original primary"
                        ]
                    )
                    > 0
                ):
                    di.NiftiName = None
                    di.ConvertImage = False

    def create_all_nii(self) -> None:
        if self.RemoveIdentifiers:
            self.anonymize()
        source_dict = defaultdict(list)
        for di in self.SeriesList:
            source_dict[di.SourcePath].append(di)
        for source_path, di_list in source_dict.items():
            if any(di.ConvertImage for di in di_list):
                logging.info("Creating Nifti for %s" % source_path)
                create_nii(self.OutputRoot / self.Metadata.dir_to_str(), source_path, di_list)
                for di in di_list:
                    if di.ConvertImage:
                        self.generate_sidecar(di)
                        if di.NiftiCreated:
                            self.generate_qa_image(di)
        self.SeriesList = sorted(
            sorted(self.SeriesList, key=lambda x: (x.StudyUID, x.SeriesNumber, x.SeriesUID)),
            key=lambda x: x.ConvertImage,
            reverse=True,
        )

    def generate_sidecar(self, di_obj: BaseInfo) -> None:
        sidecar_file = (
            self.OutputRoot / self.Metadata.dir_to_str() / "nii" / (di_obj.NiftiName + ".json")
        )
        logging.info("Writing image sidecar file to %s" % sidecar_file)
        out_dict = {k: v for k, v in self.__repr_json__().items() if k not in "SeriesList"}
        out_dict["SeriesInfo"] = di_obj
        if self.RemoveIdentifiers:
            setattr(out_dict["SeriesInfo"], "SourcePath", None)
        sidecar_file.write_text(
            json.dumps(out_dict, indent=4, sort_keys=True, cls=JSONObjectEncoder)
        )

    def generate_qa_image(self, di_obj: BaseInfo) -> None:
        qa_dir = self.OutputRoot / self.Metadata.dir_to_str() / "qa" / "conversion"
        nifti_file = (
            self.OutputRoot / self.Metadata.dir_to_str() / "nii" / (di_obj.NiftiName + ".nii.gz")
        )
        qa_file = qa_dir / (di_obj.NiftiName + ".png")
        logging.info("Creating QA image for %s" % nifti_file)
        mkdir_p(qa_dir)
        create_qa_image(nifti_file, qa_file)

    def generate_unconverted_info(self) -> None:
        info_file = (
            self.OutputRoot
            / self.Metadata.dir_to_str()
            / (self.Metadata.prefix_to_str() + "_UnconvertedInfo.json")
        )
        logging.info("Writing unconverted info file to %s" % info_file)
        out_dict = {k: v for k, v in self.__repr_json__().items() if k not in "SeriesList"}
        out_dict["SeriesList"] = [item for item in self.SeriesList if not item.NiftiCreated]
        if self.RemoveIdentifiers:
            for series in out_dict["SeriesList"]:
                series.SourcePath = None
        info_file.write_text(json.dumps(out_dict, indent=4, sort_keys=True, cls=JSONObjectEncoder))
        info_file.chmod(FILE_OCTAL)

    def anonymize(self):
        logging.info("Anonymizing info...")
        anon_study_ids = {}
        anon_series_count = {}
        for di in self.SeriesList:
            if di.StudyUID not in anon_study_ids:
                anon_study_ids[di.StudyUID] = "2.25." + str(int(str(secrets.randbits(96))))
                anon_series_count[di.StudyUID] = 0
            anon_series_count[di.StudyUID] += 1
            di.SeriesUID = anon_study_ids[di.StudyUID] + (".%03d" % anon_series_count[di.StudyUID])
            di.StudyUID = anon_study_ids[di.StudyUID]
            di.anonymize(self.DateShiftDays)
        self.LookupTable.anonymize()


def create_nii(output_dir: Path, source_path: Path, di_list: list[BaseInfo]) -> None:
    source = output_dir / source_path
    convert_di = [di for di in di_list if di.ConvertImage]
    for di in di_list:
        if not di.ConvertImage:
            di.NiftiCreated = False
            logging.info("Skipping %s" % di.SeriesUID)

    # Create the nii directory if it doesn't exist
    niidir = output_dir / "nii"
    mkdir_p(niidir)

    # If the file already exists, we won't overwrite it, just warn
    exists_list = [Path(niidir, di.NiftiName + ".nii.gz").exists() for di in convert_di]
    if any(exists_list):
        logging.warning("Naming failed for %s" % source)
        for di, exists in zip(convert_di, exists_list):
            di.NiftiCreated = False
            if exists:
                logging.warning("Attempted to create %s.nii.gz" % di.NiftiName)
                logging.warning("%s already exists." % di.NiftiName)
        logging.warning("Nifti creation failed.")
        return

    # Run dcm2niix
    dcm2niix_cmd = [
        shutil.which("dcm2niix"),
        "-b",
        "y",
        "-z",
        "y",
        "-f",
        convert_di[0].NiftiName,
        "-o",
        niidir,
        source,
    ]
    result = run(dcm2niix_cmd, stdout=PIPE, stderr=STDOUT, text=True)
    success = result.returncode == 0

    # If dcm2niix failed, remove the created files and warn
    if not success:
        logging.warning("dcm2niix failed for %s" % source)
        for di in convert_di:
            di.NiftiCreated = False
            logging.warning("Attempted to create %s.nii.gz" % di.NiftiName)
        logging.warning("dcm2niix return code: %d" % result.returncode)
        logging.warning("dcm2niix output:\n" + str(result.stdout))
        for filename in parse_dcm2niix_filenames(str(result.stdout)):
            remove_created_files(filename)
        logging.warning("Nifti creation failed.")
        return

    # Parse the output for filenames
    filenames = parse_dcm2niix_filenames(str(result.stdout))

    # Remove images that have an _e# without an EchoTime
    # Remove any images that are a, b, c, etc. of another image
    suffix_pattern = re.compile(r"^(.*)([a-z])$")
    removes = []
    for i, filename in enumerate(filenames):
        removes.append(False)
        match = suffix_pattern.match(filename.name)
        if "_e" in filename.name:
            bids_dict = json.load(open(p_add(filename, ".json")))
            if "EchoTime" not in bids_dict:
                removes[i] = True
        if not removes[i] and match:
            base, _ = match.groups()
            if base in [item.name for item in filenames]:
                removes[i] = True
        if removes[i]:
            p_add(filename, ".nii.gz").unlink()
        # Cleanup dcm2niix json files
        p_add(filename, ".json").unlink()
    filenames = [filename for filename, remove in zip(filenames, removes) if not remove]

    # Check for the right number of files
    if len(filenames) != len(convert_di):
        logging.warning(
            f"Number of dcm2niix outputs ({len(filenames)}) "
            f"did not match number expected ({len(convert_di)})."
        )
        logging.warning("dcm2niix return code: %d" % result.returncode)
        logging.warning("dcm2niix output:\n" + str(result.stdout))
        logging.warning("Nifti creation failed.")
        for di in convert_di:
            di.NiftiCreated = False
        for filename in filenames:
            remove_created_files(filename)
        return

    # Match filenames to dicom info objects
    if len(filenames) > 1:
        need_mag = any(["-MAG" in di.NiftiName for di in convert_di])
        suffixes = parse_dcm2niix_suffixes(filenames, convert_di[0].NiftiName, need_mag)
        suffixes = {suffix: filename for suffix, filename in zip(suffixes, filenames)}
        extras = ["_".join(di.NiftiName.split("_")[-1].split("-")[6:]) for di in convert_di]
        if any([sum(dyn in extra for dyn in ["POS", "INV", "DYN"]) > 1 for extra in extras]):
            logging.warning(
                "Multiple dynamic image types detected. "
                "Cannot match to filenames. Removing created files."
            )
            logging.warning("Nifti creation failed.")
            success = False
        else:
            extras = [
                tuple(sorted(extra.replace("INV", "DYN").replace("POS", "DYN").split("_")))
                for extra in extras
            ]
            filenames = [suffixes[extra] for extra in extras]

    for filename, di in zip(filenames, convert_di):
        if not success:
            break
        if len(list(filename.parent.glob(filename.name + "_Eq*.nii.gz"))) > 0:
            logging.warning("Slices missing for DICOM UID %s, not converted." % di.SeriesUID)
            logging.warning("Nifti creation failed.")
            success = False
            break
        reo_result = reorient(p_add(filename, ".nii.gz"), di.SliceOrientation)
        if not reo_result:
            success = False
            break
        if p_add(filename, "_ADC.nii.gz").exists():
            logging.info("Additional ADC images produced by dcm2niix. Removing.")
            p_add(filename, "_ADC.nii.gz").unlink()
        if filename.name != di.NiftiName:
            for file in filename.parent.glob(filename.name + ".*"):
                ext = "".join(file.suffixes)
                file.rename(niidir / (di.NiftiName + ext))

    if success:
        for di in convert_di:
            if (niidir / (di.NiftiName + ".nii.gz")).exists():
                di.NiftiCreated = True
                di.NiftiHash = hash_file_dir(niidir / (di.NiftiName + ".nii.gz"))
                logging.info(
                    "Nifti created successfully at %s" % (niidir / (di.NiftiName + ".nii.gz"))
                )
            else:
                di.NiftiCreated = False
                logging.warning("Nifti file missing: %s.nii.gz" % di.NiftiName)
                logging.warning("Nifti creation failed.")
    else:
        for di in convert_di:
            di.NiftiCreated = False
        for filename in filenames:
            remove_created_files(filename)


class TruncatedImageValue:
    def __init__(
        self, value: Optional[Union[List[float], Tuple[float], np.ndarray]], precision: float = 1e-4
    ) -> None:
        self.value = value
        self.precision = precision
        self.length = int(abs(np.log10(self.precision)))

    def __eq__(self, other: Union[Any, TruncatedImageValue]) -> bool:
        if isinstance(other, TruncatedImageValue):
            if self.value is None:
                return other.value is None
            return all(
                [
                    abs(self.value[i] - other.value[i]) <= self.precision
                    for i in range(len(self.value))
                ]
            )
        else:
            return super().__eq__(other)

    def __hash__(self) -> int:
        return hash(tuple(self.truncate())) if self.value is not None else hash(None)

    def __getitem__(self, item) -> Optional[float]:
        return self.value[item] if self.value is not None else None

    def __repr_json__(self) -> Optional[List[float]]:
        return self.truncate()

    def truncate(self) -> Optional[List[float]]:
        return (
            [float(np.around(num / 2, self.length) * 2) for num in self.value]
            if self.value is not None
            else None
        )


class ImageOrientation(TruncatedImageValue):
    def get_plane(self) -> Optional[str]:
        if self.value is None:
            return None
        if len(self.value) == 6:  # Direction Cosines
            plane = int(np.argmax([abs(x) for x in np.cross(self.value[0:3], self.value[3:6])]))
            return DCM_ORIENT_PLANES.get(plane, None)
        if len(self.value) == 4:  # Angulation + Plane
            return PARREC_ORIENTATIONS.get(self.value[3], None)



import io
from pathlib import Path

import matplotlib.pyplot as plt
import nibabel as nib
import numpy as np
import trimesh
from PIL import Image
from radifox.utils.resize.scipy import resize

from .utils import get_tkr_matrix

BINARY_LUT = {1: [255, 0, 0]}

BRAIN_COLOR_LUT = {
    4: [0, 15, 247],
    11: [255, 245, 9],
    23: [178, 178, 22],
    30: [255, 255, 98],
    31: [131, 178, 46],
    32: [208, 255, 123],
    35: [255, 9, 245],
    36: [178, 178, 126],
    37: [255, 255, 203],
    38: [178, 96, 81],
    39: [255, 173, 158],
    40: [178, 142, 36],
    41: [255, 218, 112],
    44: [89, 89, 89],
    45: [165, 165, 165],
    47: [146, 178, 32],
    48: [222, 255, 108],
    49: [0, 61, 147],
    50: [76, 137, 224],
    51: [0, 139, 108],
    52: [76, 215, 185],
    53: [190, 0, 0],
    54: [255, 0, 0],
    55: [178, 54, 0],
    56: [255, 131, 76],
    57: [178, 124, 0],
    58: [255, 200, 76],
    59: [173, 0, 0],
    60: [250, 76, 76],
    61: [37, 0, 0],
    62: [113, 76, 76],
    71: [255, 34, 221],
    72: [255, 62, 192],
    73: [255, 91, 163],
    75: [188, 76, 76],
    76: [111, 0, 0],
    100: [102, 178, 75],
    101: [179, 255, 152],
    102: [106, 0, 0],
    103: [182, 76, 76],
    104: [0, 129, 178],
    105: [76, 206, 255],
    106: [34, 178, 143],
    107: [111, 255, 220],
    108: [174, 178, 4],
    109: [250, 255, 80],
    112: [178, 31, 0],
    113: [255, 107, 76],
    114: [178, 155, 0],
    115: [255, 231, 76],
    116: [151, 0, 0],
    117: [227, 76, 76],
    118: [0, 68, 178],
    119: [76, 144, 255],
    120: [0, 0, 170],
    121: [76, 76, 246],
    122: [135, 0, 0],
    123: [212, 76, 76],
    124: [0, 0, 112],
    125: [76, 76, 188],
    128: [178, 141, 0],
    129: [255, 218, 76],
    132: [166, 0, 0],
    133: [243, 76, 76],
    134: [178, 168, 0],
    135: [255, 245, 76],
    136: [0, 143, 178],
    137: [76, 220, 255],
    138: [89, 178, 89],
    139: [165, 255, 165],
    140: [0, 0, 97],
    141: [76, 76, 173],
    142: [0, 3, 178],
    143: [76, 79, 255],
    144: [178, 100, 0],
    145: [255, 176, 76],
    146: [0, 115, 178],
    147: [76, 192, 255],
    148: [0, 168, 178],
    149: [76, 245, 255],
    150: [0, 15, 178],
    151: [76, 92, 255],
    152: [0, 0, 127],
    153: [76, 76, 204],
    154: [178, 89, 0],
    155: [255, 165, 76],
    156: [160, 178, 17],
    157: [237, 255, 94],
    160: [178, 127, 0],
    161: [255, 204, 76],
    162: [0, 85, 178],
    163: [76, 162, 255],
    164: [0, 42, 178],
    165: [76, 119, 255],
    166: [76, 178, 101],
    167: [153, 255, 178],
    168: [0, 155, 178],
    169: [76, 232, 255],
    170: [119, 0, 0],
    171: [196, 76, 76],
    172: [94, 0, 0],
    173: [171, 76, 76],
    174: [63, 178, 115],
    175: [139, 255, 191],
    176: [48, 178, 129],
    177: [125, 255, 206],
    178: [0, 101, 178],
    179: [76, 178, 255],
    180: [178, 66, 0],
    181: [255, 142, 76],
    182: [0, 30, 178],
    183: [76, 107, 255],
    184: [178, 14, 0],
    185: [255, 90, 76],
    186: [117, 178, 61],
    187: [193, 255, 137],
    190: [0, 0, 157],
    191: [76, 76, 233],
    192: [0, 0, 143],
    193: [76, 76, 219],
    194: [20, 178, 158],
    195: [96, 255, 234],
    196: [178, 114, 0],
    197: [255, 190, 76],
    198: [5, 178, 173],
    199: [81, 255, 249],
    200: [178, 79, 0],
    201: [255, 156, 76],
    202: [178, 1, 0],
    203: [255, 77, 76],
    204: [0, 53, 178],
    205: [76, 130, 255],
    206: [178, 48, 0],
    207: [255, 125, 76],
}

COLORBLIND_LUT = {
    1: [229, 159, 1],
    2: [86, 180, 232],
    3: [0, 159, 115],
    4: [240, 228, 55],
    5: [0, 114, 177],
    6: [213, 94, 0],
    7: [204, 121, 167],
    8: [160, 111, 1],
    9: [60, 126, 162],
    10: [0, 111, 81],
    11: [168, 160, 46],
    12: [0, 80, 124],
    13: [149, 66, 0],
    14: [143, 85, 117],
}

luts = {"binary": BINARY_LUT, "brain_color": BRAIN_COLOR_LUT, "colorblind": COLORBLIND_LUT}


def create_montage(img_obj, axial_slices, coronal_slices, sagittal_slices):
    reslicer = Reslicer(img_obj)
    max_size = max(reslicer.data.shape)
    slices = [[], [], []]
    for i, (plane, dist_list) in enumerate(
        zip(["axial", "sagittal", "coronal"], [axial_slices, sagittal_slices, coronal_slices])
    ):
        slice_list = np.around(np.array(dist_list) * reslicer.get_num_slices(plane))
        for slice_num in slice_list:
            sel_slice = reslicer.get_slice(int(slice_num), plane)
            pad_width = [
                (
                    int(np.floor((max_size - sel_slice.shape[i]) / 2.0)),
                    int(np.ceil((max_size - sel_slice.shape[i]) / 2.0)),
                )
                for i in range(2)
            ]
            slices[i].append(np.pad(sel_slice, pad_width, "constant"))
    return np.concatenate([np.concatenate(slices[i], 0) for i in range(len(slices))], 1)


def create_qa_image(
    input_filename,
    output_filename,
    overlay_filename=None,
    overlay_lut="binary",
    axial_slices=(0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7),
    coronal_slices=(0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7),
    sagittal_slices=(0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7),
    alpha=0.4,
):
    input_obj = nib.load(input_filename)
    base_img = create_montage(input_obj, axial_slices, coronal_slices, sagittal_slices)
    base_img -= np.min(base_img)
    base_img = np.array(base_img / np.percentile(base_img, 99.9) * 255.0)
    base_img[base_img > 255.0] = 255.0
    base_img = base_img.astype(np.ubyte).T

    if overlay_filename is not None:
        overlay_obj = nib.load(overlay_filename)
        overlay_img = (
            create_montage(overlay_obj, axial_slices, coronal_slices, sagittal_slices)
            .astype(np.ubyte)
            .T
        )
        lut = np.zeros((256, 3))
        for k, v in luts[overlay_lut].items():
            lut[k, :] = v
        colored_overlay = np.zeros(overlay_img.shape + (3,))
        for i in range(3):
            colored_overlay[overlay_img > 0, i] = np.take(lut[:, i], overlay_img[overlay_img > 0])
        base_img = np.repeat(base_img[:, :, np.newaxis], 3, axis=2)
        base_img = ((colored_overlay * alpha) + (base_img * (1 - alpha))).astype(np.ubyte)

    Image.fromarray(base_img).save(output_filename)


class Reslicer:
    """
    Reslice a NIfTI image

    This class reslices and image to a set resolution and provides as way to
    extract a slice in a specific orientation. It can currently bring the
    image to axial, coronal, and sagittal views.

    Params:
        data (np.ndarray): The Nifti image object from nibabel
        orient (tuple[str]): orientation tuple
    """

    inplane_dirs = {
        "axial": (("L", "P"), ("R", "A")),
        "sagittal": (("P", "I"), ("A", "S")),
        "coronal": (("L", "I"), ("R", "S")),
    }
    slice_dirs = {"axial": "S", "sagittal": "L", "coronal": "P"}
    flip_dict = {"L": "R", "R": "L", "P": "A", "A": "P", "I": "S", "S": "I"}

    def __init__(self, nii_obj, vox_res=1.0, order=0):
        """
        Args:
            nii_obj (nib.Nifti1Image): The Nifti image object from nibabel
            vox_res (float): Resliced object voxel resolution
            order (int, optional): The interpolation order (0 for nearest
                interpolation, 1 for linear interpolation, etc.)
        """
        # noinspection PyTypeChecker
        obj_3d = nii_obj if len(nii_obj.shape) == 3 else nib.four_to_three(nii_obj)[0]
        data = obj_3d.get_fdata()
        self.data: np.ndarray = resize(
            data, [vox_res / s for s in obj_3d.header.get_zooms()], order=order
        )
        self.orient = nib.aff2axcodes(nii_obj.affine)

    def get_num_slices(self, plane="axial") -> int:
        if self.slice_dirs[plane] in self.orient:
            return self.data.shape[self.orient.index(self.slice_dirs[plane])]
        return self.data.shape[self.orient.index(self.flip_dict[self.slice_dirs[plane]])]

    def get_slice(self, slice_num, plane="axial") -> np.ndarray:
        slices, dirs = [], []
        for i, code in enumerate(self.orient):
            if code in self.inplane_dirs[plane][0]:
                slices.append(slice(None, None))
                dirs.append(code)
            elif code in self.inplane_dirs[plane][1]:
                slices.append(slice(None, None, -1))
                dirs.append(self.flip_dict[code])
            else:
                if code == self.slice_dirs[plane]:
                    slices.append(slice(slice_num, slice_num + 1))
                else:
                    slices.append(
                        slice(self.data.shape[i] - slice_num - 1, self.data.shape[i] - slice_num)
                    )
        sel_slice = np.squeeze(self.data[tuple(slices)])
        return sel_slice if tuple(dirs) == self.inplane_dirs[plane][0] else sel_slice.T


def create_surface_qa_image(
    surf_file,
    img_file,
    output_file,
    color="red",
    axial_slices=(0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7),
    coronal_slices=(0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7),
    sagittal_slices=(0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7),
):
    linewidth = 0.5
    if color == "binary":
        linecolor = "red"
    if isinstance(color, (list, tuple)):
        linecolor = color[0]
        linewidth = color[1]
    else:
        linecolor = color
    img_obj = nib.Nifti1Image.load(img_file)
    img_data = img_obj.get_fdata()
    # noinspection PyUnresolvedReferences
    tk = get_tkr_matrix(img_obj.shape, img_obj.header.get_zooms())

    surf_obj = nib.GiftiImage.load(surf_file)
    mesh = trimesh.Trimesh(*surf_obj.agg_data(("pointset", "triangle")))
    # noinspection PyUnresolvedReferences
    mesh.vertices = nib.affines.apply_affine(np.linalg.inv(tk), mesh.vertices)

    directions = ["axial", "sagittal", "coronal"]
    axes = {"sagittal": 0, "coronal": 1, "axial": 2}
    slice_levels = {"axial": axial_slices, "sagittal": sagittal_slices, "coronal": coronal_slices}

    slice_arrs = []
    for direction in directions:
        size = img_data.shape[axes[direction]]
        slices = np.rint(np.array(slice_levels[direction]) * size).astype(int)
        for i, slice_idx in enumerate(slices):
            if direction == "sagittal":
                slice_arrs.append(img_data[slice_idx, :, ::-1].ravel())
            elif direction == "coronal":
                slice_arrs.append(img_data[:, slice_idx, ::-1].ravel())
            else:  # 'axial'
                slice_arrs.append(img_data[:, :, slice_idx].ravel())
    slice_arr = np.concatenate(slice_arrs, axis=0)
    img_data -= slice_arr.min()
    img_data = np.array(img_data / np.percentile(slice_arr, 99.9) * 255.0)
    img_data[img_data > 255.0] = 255.0
    img_data = img_data.astype(np.ubyte)

    imgs = []
    for d, direction in enumerate(directions):
        imgs.append([])
        size = img_data.shape[axes[direction]]
        slices = np.rint(np.array(slice_levels[direction]) * size).astype(int)
        if direction == "sagittal":
            plane_normal = [1, 0, 0]
            adding = [192, 0]
        elif direction == "coronal":
            plane_normal = [0, 1, 0]
            adding = [192, 176]
        else:
            plane_normal = [0, 0, 1]
            adding = [16, 0]
        sections = mesh.section_multiplane(
            plane_origin=[0, 0, 0],
            plane_normal=plane_normal,
            heights=slices,
        )

        for i, slice_idx in enumerate(slices):
            fig = plt.figure(frameon=False, figsize=(224 / 72, 224 / 72), dpi=72)
            ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
            ax.set_axis_off()
            fig.add_axes(ax)

            if direction == "sagittal":
                slice_data = img_data[slice_idx, :, ::-1]
                extent = (0, img_data.shape[1], 16, img_data.shape[2] + 16)
            elif direction == "coronal":
                slice_data = img_data[:, slice_idx, ::-1]
                extent = (16, img_data.shape[0] + 16, 16, img_data.shape[2] + 16)
            else:  # 'axial'
                slice_data = img_data[:, :, slice_idx]
                extent = (16, img_data.shape[0] + 16, 0, img_data.shape[1])

            # Plot the slice with the correct extent
            ax.imshow(np.zeros((224, 224)), extent=(0, 224, 0, 224), cmap="gray")
            ax.imshow(slice_data.T, extent=extent, cmap="gray")

            # Find intersections
            if sections[i] is not None:
                # noinspection PyTypeChecker
                for points in sections[i].discrete:
                    points[:, 0] += adding[0]
                    points[:, 1] += adding[1]
                    if direction in ["sagittal", "coronal"]:
                        points = points[:, ::-1]
                        points[:, 1] = img_data.shape[2] - points[:, 1] + 16
                        if direction == "coronal":
                            points[:, 0] = img_data.shape[0] - points[:, 0]
                    else:
                        points[:, 1] = img_data.shape[1] - points[:, 1]
                    ax.plot(*points.T, color=linecolor, linewidth=linewidth)
            buf = io.BytesIO()
            fig.savefig(buf, format="png")
            buf.seek(0)
            imgs[d].append(np.array(Image.open(buf)))
            buf.close()
            plt.close(fig)
    montage = np.concatenate([np.concatenate(img_dir, 1) for img_dir in imgs], 0)
    montage_im = Image.fromarray(montage)
    montage_im.save(output_file)



import datetime
import logging
from pathlib import Path
import sys

from .utils import mkdir_p

WARNING_DEBUG = 25


class LogFilter(logging.Filter):
    def __init__(self, level: int) -> None:
        super().__init__()
        self.level = level

    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno <= self.level


def create_loggers(
    log_dir: Path,
    log_prefix: str,
    verbose: bool = False,
    add_stream_handler: bool = True,
) -> None:
    mkdir_p(log_dir)
    log_formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s")
    timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    # noinspection PyTypeChecker
    log_file = logging.FileHandler(log_dir / f"{log_prefix}-{timestamp}-info.log", delay=True)
    log_file.setFormatter(log_formatter)
    logging.getLogger().setLevel(logging.DEBUG if verbose else logging.INFO)
    logging.getLogger().addHandler(log_file)
    if add_stream_handler:
        log_stream = logging.StreamHandler(sys.stdout)
        log_stream.setFormatter(log_formatter)
        logging.getLogger().addHandler(log_stream)
    # noinspection PyTypeChecker
    warn_file = logging.FileHandler(log_dir / f"{log_prefix}-{timestamp}-warning.log", delay=True)
    warn_file.addFilter(LogFilter(logging.ERROR - 1))
    logging.addLevelName(WARNING_DEBUG, "WARNING-DEBUG")
    warn_file.setFormatter(log_formatter)
    warn_file.setLevel(WARNING_DEBUG if verbose else logging.WARNING)
    logging.getLogger().addHandler(warn_file)
    # noinspection PyTypeChecker
    error_file = logging.FileHandler(log_dir / f"{log_prefix}-{timestamp}-error.log", delay=True)
    error_file.addFilter(LogFilter(logging.ERROR))
    error_file.setFormatter(log_formatter)
    error_file.setLevel(logging.ERROR)
    logging.getLogger().addHandler(error_file)



from codecs import BOM_UTF8
from collections.abc import Sequence
import csv
from datetime import datetime, date, time, timedelta
import hashlib
import logging
import os
from pathlib import Path
import re
import shutil
from subprocess import check_output

import nibabel as nib
from pydicom.dataset import Dataset, FileDataset, Tag

ORIENT_CODES = {"sagittal": "PIL", "coronal": "LIP", "axial": "LPS"}


# http://stackoverflow.com/a/22718321
def mkdir_p(path: Path, mode: int = 0o777) -> None:
    path.mkdir(mode=mode, parents=True, exist_ok=True)


def copytree_link(source: Path, dest: Path, method: str) -> None:
    if method in ["hardlink", "softlink", "copy"]:
        func = {"hardlink": os.link, "symlink": os.symlink, "copy": shutil.copyfile}[method]
    else:
        raise ValueError(f"Unsupported copy method: {method}")
    dest.mkdir(parents=True, exist_ok=True)
    for path in source.glob("*"):
        if path.is_file():
            func(path, dest / path.name)
        elif path.is_dir():
            (dest / path.name).mkdir()
            copytree_link(path, dest / path.name, method)


# http://stackoverflow.com/a/10840586
def silentremove(filename: Path) -> None:
    import shutil
    import errno

    try:
        if filename.is_file():
            filename.unlink()
        else:
            shutil.rmtree(filename)
    except OSError as e:  # this would be "except OSError, e:" before Python 2.6
        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise  # re-raise exception if a different error occurred


def read_csv(csv_filename: Path) -> dict[str, list[str]]:
    data = csv_filename.read_bytes()
    codec = "UTF-8-SIG" if data.startswith(BOM_UTF8) else "UTF-8"
    data = data.decode(codec)
    line_sep = "\r\n" if "\r\n" in data else "\n"
    data = data.split(line_sep)
    out_dict = {key: [] for key in data[0].split(",")}
    for row in csv.DictReader(data):
        for key in out_dict.keys():
            out_dict[key].append(row[key])
    return out_dict


def convert_dicom_date(date_str: str) -> date:
    return datetime.strptime(date_str.replace("-", ""), "%Y%m%d").date()


def convert_dicom_time(time_str: str) -> time:
    return datetime.strptime(time_str.split(".")[0].ljust(6, "0"), "%H%M%S").time()


def convert_dicom_datetime(dt_str: str) -> date:
    return datetime.strptime(dt_str.replace("-", "").split(".")[0], "%Y%m%d%H%M%S")


vr_corr = {
    "CS": str,
    "DA": convert_dicom_date,
    "DS": float,
    "DT": convert_dicom_datetime,
    "FD": float,
    "FL": float,
    "IS": int,
    "LO": str,
    "LT": str,
    "OB": bytes,
    "PN": str,
    "SH": str,
    "SL": int,
    "SS": int,
    "ST": str,
    "TM": convert_dicom_time,
    "UI": str,
    "UL": int,
    "US": int,
}


def extract_de(
    ds: Dataset | FileDataset,
    label: str,
    series_uid,
    keep_list: bool = False,
) -> None | tuple | float | int | str:
    if label not in ds:
        return tuple() if keep_list else None
    de = ds[label]
    if de.VM == 0:
        return tuple() if keep_list else None
    value = [de.value] if de.VM == 1 else de.value
    try:
        out_list = []
        for item in value:
            out_list.append(vr_corr[de.VR](item))
        out_list = tuple(out_list)
    except ValueError:
        logging.warning(
            "Data element (%s) will not conform to required type (%s) for %s."
            % (de.name, str(vr_corr[de.VR]), series_uid)
        )
        return tuple() if keep_list else None
    return out_list[0] if (len(out_list) == 1 and not keep_list) else out_list


def is_intstr(test_str: str) -> bool:
    try:
        int(test_str)
    except ValueError:
        return False
    return True


def reorient(input_file: Path, orientation: str) -> bool:
    # noinspection PyTypeChecker
    input_obj = nib.Nifti1Image.load(input_file)
    input_orient = "".join(nib.aff2axcodes(input_obj.affine))
    target_orient = ORIENT_CODES[orientation]
    logging.debug("Reorienting %s from %s to %s." % (input_file, input_orient, target_orient))
    if input_orient != tuple(target_orient):
        try:
            orig_ornt = nib.orientations.io_orientation(input_obj.affine)
            targ_ornt = nib.orientations.axcodes2ornt(target_orient)
            ornt_xfm = nib.orientations.ornt_transform(orig_ornt, targ_ornt)

            affine = input_obj.affine.dot(nib.orientations.inv_ornt_aff(ornt_xfm, input_obj.shape))
            data = nib.orientations.apply_orientation(input_obj.dataobj, ornt_xfm)
            nib.Nifti1Image(data, affine, input_obj.header).to_filename(str(input_file))
            return True
        except ValueError:
            logging.warning("Reorientation failed for %s" % input_file)
            return False


def allowed_archives() -> (list[str], list[str]):
    allowed_exts = []
    allowed_names = []
    for names, extensions, _ in shutil.get_unpack_formats():
        allowed_exts.extend(extensions)
        allowed_names.append(names)
    if ".zip" in allowed_exts:
        allowed_exts.append(".zip.zip")
    return allowed_names, allowed_exts


def extract_archive(input_zipfile: Path, output_dir: Path) -> None:
    logging.info("Extracting archive")
    mkdir_p(output_dir)
    # noinspection PyTypeChecker
    shutil.unpack_archive(input_zipfile, output_dir)
    logging.info("Extraction complete")


def make_tuple(item: bytes | str | Sequence) -> tuple:
    if isinstance(item, (bytes, str)):
        return tuple([item])
    return tuple(item) if isinstance(item, Sequence) else tuple([item])


def remove_created_files(filename: Path) -> None:
    for imgname in [
        f
        for f in filename.parent.glob(filename.name + "*")
        if re.search(filename.name + r"_*[A-Za-z0-9_]*\..+$", f.name)
    ]:
        imgname.unlink()


def parse_dcm2niix_filenames(stdout: str) -> list[Path]:
    filenames = []
    for line in stdout.split("\n"):
        if line.startswith("Convert "):  # output
            fname = Path(str(re.search(r"\S+/\S+", line).group(0)))
            filenames.append(fname.resolve())
    return filenames


def add_acq_num(name: str, count: int) -> str:
    prefix = "_".join(name.split("_")[:-1])
    contrast_arr = name.split("_")[-1].split("-")
    addons = "" if len(contrast_arr) == 6 else ("-" + "-".join(contrast_arr[6:]))
    base_contrast = "-".join(contrast_arr[:6])
    return prefix + "_" + base_contrast + ("-ACQ%d" % count) + addons


FILE_OCTAL = 0o660
DIR_OCTAL = 0o2770


def has_permissions(path: Path, octal: int = DIR_OCTAL) -> bool:
    return int("0o" + oct(path.stat().st_mode)[-4:], 8) == octal


def recursive_chmod(
    directory: Path,
    dir_octal: int = DIR_OCTAL,
    file_octal: int = FILE_OCTAL,
) -> None:
    if not directory.exists():
        return
    elif directory.is_file():
        directory.chmod(file_octal)
    else:
        directory.chmod(dir_octal)
        for item in directory.rglob("*"):
            if item.is_dir():
                item.chmod(dir_octal)
            elif item.is_file():
                item.chmod(file_octal)


def find_closest(target: int, to_check: list[int]) -> int | None:
    if len(to_check) < 1:
        return None
    elif len(to_check) == 1:
        return to_check[0]
    signed_dists = []
    for check_val in to_check:
        signed_dists.append((check_val - target, check_val))
    min_dist = min([abs(val[0]) for val in signed_dists])
    candidates = [val[1] for val in signed_dists if abs(val[0]) == min_dist]
    return candidates[0] if len(candidates) == 1 else min(candidates)


def hash_file(
    filename: Path,
    include_names: bool = True,
    hashfunc: str = "sha256",
    *,
    _bufsize=2**18,
) -> str:
    hashobj = hashlib.new(hashfunc)
    if include_names:
        hashobj.update(filename.name.encode())
    with filename.open("rb") as fp:
        buf = bytearray(_bufsize)  # Reusable buffer to reduce allocations.
        view = memoryview(buf)
        while True:
            # noinspection PyUnresolvedReferences
            size = fp.readinto(buf)
            if size == 0:
                break  # EOF
            hashobj.update(view[:size])
    return str(hashobj.hexdigest())


def hash_dir(directory, include_names: bool = True, hashfunc: str = "sha256") -> str:
    hashobj = hashlib.new(hashfunc)
    for path in sorted(directory.iterdir(), key=lambda p: str(p).lower()):
        hashobj.update(hash_file_dir(path, include_names, hashfunc).encode())
    if include_names:
        hashobj.update(directory.name.encode())
    return str(hashobj.hexdigest())


def hash_file_dir(file_dir: Path, include_names: bool = True, hashfunc: str = "sha256") -> str:
    if file_dir.is_file():
        return hash_file(file_dir, include_names=include_names, hashfunc=hashfunc)
    elif file_dir.is_dir():
        return hash_dir(file_dir, include_names=include_names, hashfunc=hashfunc)


def hash_file_list(
    file_list: list[Path], include_names: bool = True, hashfunc: str = "sha256"
) -> str:
    hashobj = hashlib.new(hashfunc)
    for path in file_list:
        hashobj.update(hash_file(path, include_names, hashfunc).encode())
    return str(hashobj.hexdigest())


def hash_value(value: str | None, hashfunc: str = "sha256") -> str | None:
    if value is None:
        return None
    m = hashlib.new(hashfunc)
    m.update(value.encode())
    return str(m.hexdigest())


def p_add(path: Path, extra: str) -> Path:
    return path.parent / (path.name + extra)


def get_software_versions() -> dict[str, str]:
    dcm2niix_version = (
        check_output("dcm2niix --version; exit 0", shell=True)
        .decode()
        .strip()
        .split("\n")[-1]
        .strip()
    )
    return {"dcm2niix": dcm2niix_version}


def version_check(saved_version: str, current_version: str) -> bool:
    if "dev" in saved_version or "dev" in current_version:
        return False
    saved_arr = saved_version.split("-")[0].split(".")
    current_arr = current_version.split("-")[0].split(".")
    for saved, current in zip(saved_arr, current_arr):
        if int(saved) < int(current):
            return False
    return True


def shift_date(datetime_str: str, date_shift_days: int = 0) -> str:
    orig_date = datetime.strptime(datetime_str, "%Y-%m-%d %H:%M:%S")
    return (orig_date + timedelta(days=date_shift_days)).strftime("%Y-%m-%d %H:%M:%S")


def none_to_float(value: float | None) -> float:
    return float("inf") if value is None else value


def get_flattened_dataset(dataset: FileDataset | Dataset) -> Dataset:
    return Dataset({de.tag: de for de in dataset.iterall() if de.VR != "SQ"})


EXCLUDE_TAGS = [
    Tag("SharedFunctionalGroupsSequence"),
    Tag("PerFrameFunctionalGroupsSequence"),
    Tag("DimensionIndexSequence"),
    Tag("NumberOfFrames"),
    Tag("SourceImageEvidenceSequence"),
    Tag("ReferencedImageEvidenceSequence"),
    Tag("PixelData"),
]


def fix_sf_headers(dataset: Dataset) -> Dataset:
    if "EffectiveEchoTime" in dataset:
        dataset.EchoTime = dataset.EffectiveEchoTime
    scan_seq: list = (
        (
            dataset.ScanningSequence
            if dataset["ScanningSequence"].VM > 1
            else [dataset.ScanningSequence]
        )
        if "ScanningSequence" in dataset
        else []
    )
    if "EchoPulseSequence" in dataset:
        if dataset.EchoPulseSequence != "SPIN":
            scan_seq.append("GR")
        if dataset.EchoPulseSequence != "GRADIENT":
            scan_seq.append("SE")
    if dataset.get("InversionRecovery", "NO") == "YES":
        scan_seq.append("IR")
    if dataset.get("EchoPlanarPulseSequence", "NO") == "YES":
        scan_seq.append("EP")
    dataset.ScanningSequence = sorted(set(scan_seq))

    seq_var: list = (
        (
            dataset.SequenceVariant
            if dataset["SequenceVariant"].VM > 1
            else [dataset.SequenceVariant]
        )
        if "SequenceVariant" in dataset
        else []
    )
    if dataset.get("SegmentedKSpaceTraversal", "SINGLE") != "SINGLE":
        seq_var.append("SK")
    if dataset.get("MagnetizationTransfer", "NONE") != "NONE":
        seq_var.append("MTC")
    if dataset.get("SteadyStatePulseSequence", "NONE") != "NONE":
        seq_var.append("TRSS" if dataset.SteadyStatePulseSequence == "TIME_REVERSED" else "SS")
    if dataset.get("Spoiling", "NONE") != "NONE":
        seq_var.append("SP")
    if dataset.get("OversamplingPhase", "NONE") != "NONE":
        seq_var.append("OSP")
    if len(seq_var) == 0:
        seq_var.append("NONE")
    dataset.SequenceVariant = sorted(set(seq_var))

    scan_opts: list = (
        (dataset.ScanOptions if dataset["ScanOptions"].VM > 1 else [dataset.ScanOptions])
        if "ScanOptions" in dataset
        else []
    )
    if dataset.get("RectilinearPhaseEncodeReordering", "LINEAR") != "LINEAR":
        scan_opts.append("PER")
    frame_type3 = dataset.FrameType[2]
    if frame_type3 == "ANGIO":
        dataset.AngioFlag = "Y"
    if frame_type3.startswith("CARD"):
        scan_opts.append("CG")
    if frame_type3.endswith("RESP_GATED"):
        scan_opts.append("RG")
    if "PartialFourierDirection" in dataset:
        if dataset.PartialFourierDirection == "PHASE":
            scan_opts.append("PFP")
        elif dataset.PartialFourierDirection == "FREQUENCY":
            scan_opts.append("PFF")
    if dataset.get("SpatialPresaturation", "NONE") != "NONE":
        scan_opts.append("SP")
    if dataset.get("SpectrallySelectedSuppression", "NONE").startswith("FAT"):
        scan_opts.append("FS")
    if dataset.get("FlowCompensation", "NONE") != "NONE":
        scan_opts.append("FC")
    dataset.ScanOptions = sorted(set(scan_opts))
    return dataset


def create_sf_headers(dataset: Dataset | FileDataset) -> list[Dataset]:
    shared_ds = Dataset({de.tag: de for de in dataset if de.tag not in EXCLUDE_TAGS})
    shared_ds.file_meta = dataset.file_meta
    shared_ds.update(get_flattened_dataset(dataset.SharedFunctionalGroupsSequence[0]))
    flattened_frame_ds_list = [
        get_flattened_dataset(dataset.PerFrameFunctionalGroupsSequence[i])
        for i in range(len(dataset.PerFrameFunctionalGroupsSequence))
    ]

    sf_ds_list = []
    for flat_frame_ds in flattened_frame_ds_list:
        sf_ds = Dataset({de.tag: de for de in shared_ds})
        sf_ds.update(flat_frame_ds)
        sf_ds_list.append(fix_sf_headers(sf_ds))
    return sf_ds_list


def parse_dcm2niix_suffixes(filenames: list[Path], base: str, add_mag: bool) -> list[tuple[str]]:
    suffixes = [filename.name.replace(base, "") for filename in filenames]
    for i in range(len(suffixes)):
        suffixes[i] = suffixes[i].replace("_e", "_ECHO")
        suffixes[i] = suffixes[i].replace("_ph", "_PHA")
        suffixes[i] = suffixes[i].replace("_real", "_REA")
        suffixes[i] = suffixes[i].replace("_imaginary", "_IMA")
        suffixes[i] = suffixes[i].replace("_t", "_DYN")
    if add_mag or any([cplx in suffix for suffix in suffixes for cplx in ["PHA", "REA", "IMA"]]):
        for i in range(len(suffixes)):
            if not any([cplx in suffixes[i] for cplx in ["PHA", "REA", "IMA"]]):
                suffixes[i] += "_MAG"
    # Replace numbers after DYN with 1,2,3... in order of original number
    dyn_nums = sorted(
        {
            int(re.search(r"_DYN(\d+)", suffix).group(1))
            for suffix in suffixes
            if re.search(r"_DYN(\d+)", suffix) is not None
        }
    )
    for i in range(len(suffixes)):
        if re.search(r"_DYN(\d+)", suffixes[i]) is not None:
            dyn_idx = dyn_nums.index(int(re.search(r"_DYN(\d+)", suffixes[i]).group(1))) + 1
            suffixes[i] = re.sub(r"_DYN(\d+)", "_DYN%d" % dyn_idx, suffixes[i])
    # Do the same for echoes
    echo_nums = sorted(
        {
            int(re.search(r"_ECHO(\d+)", suffix).group(1))
            for suffix in suffixes
            if re.search(r"_ECHO(\d+)", suffix) is not None
        }
    )
    for i in range(len(suffixes)):
        if re.search(r"_ECHO(\d+)", suffixes[i]) is not None:
            dyn_idx = echo_nums.index(int(re.search(r"_ECHO(\d+)", suffixes[i]).group(1))) + 1
            suffixes[i] = re.sub(r"_ECHO(\d+)", "_ECHO%d" % dyn_idx, suffixes[i])
    return [tuple(sorted(suffix[1:].split("_"))) for suffix in suffixes]



import argparse
import os

from gunicorn.app.wsgiapp import WSGIApplication


class StandaloneApplication(WSGIApplication):
    def __init__(self, app_uri, options=None):
        self.options = options or {}
        self.app_uri = app_uri
        super().__init__()

    def load_config(self):
        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)


def run(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default="localhost")
    parser.add_argument("--port", default="5000")
    parser.add_argument("--root-directory")
    parser.add_argument("--secret-key")
    parser.add_argument("--workers", type=int, default=1)
    parsed = parser.parse_args(args)

    os.environ["QA_HOST"] = parsed.host
    os.environ["QA_PORT"] = parsed.port
    if parsed.root_directory is not None:
        os.environ["QA_DATA_DIR"] = parsed.root_directory
    if parsed.secret_key is not None:
        os.environ["QA_SECRET_KEY"] = parsed.secret_key

    options = {
        "bind": "%s:%s" % (parsed.host, parsed.port),
        "workers": parsed.workers,
    }
    StandaloneApplication("radifox.qa.app:app", options).run()



from datetime import datetime
import logging
from pathlib import Path
import re
import secrets
import shutil
from typing import Optional

import numpy as np

from .base import BaseInfo, BaseSet, ImageOrientation, TruncatedImageValue
from .lut import LookupTable
from .metadata import Metadata
from .nib_parrec_fork import PARRECHeader, PARRECImage, TruncatedPARRECError
from .parrec_writer import split_fix_parrec
from .utils import silentremove

GENERAL_INFO_FIELDS = {
    "StudyDescription": "exam_name",
    "SeriesDescription": "protocol_name",
    "SeriesNumber": "acq_nr",
    "SequenceName": "tech",
    "AcquisitionDimension": "scan_mode",
}
COMPLEX_IMAGE_TYPES = {0: "MAGNITUDE", 1: "REAL", 2: "IMAGINARY", 3: "PHASE"}


class ParrecInfo(BaseInfo):
    def __init__(self, par_file: Path, manual_args: Optional[dict] = None) -> None:
        super().__init__(par_file)
        file_map = PARRECImage.filespec_to_file_map(str(par_file))
        with file_map["header"].get_prepare_fileobj("rt") as hdr_fobj:
            try:
                hdr = PARRECHeader.from_fileobj(hdr_fobj, permit_truncated=False)
                self.Truncated = False
            except TruncatedPARRECError:
                hdr = PARRECHeader.from_fileobj(hdr_fobj, permit_truncated=True)
                self.Truncated = True

        self.SeriesUID = par_file.name[:-4]
        self.StudyUID = ".".join(self.SeriesUID.split(".")[:3])
        self.Manufacturer = "PHILIPS"
        for key, value in GENERAL_INFO_FIELDS.items():
            setattr(self, key, hdr.general_info[value])
        self.ReconstructionNumber = hdr.general_info["recon_nr"]
        self.MTContrast = hdr.general_info["mtc"]
        self.EPIFactor = hdr.general_info["epi_factor"]
        self.DiffusionFlag = hdr.general_info["diffusion"]
        self.AcquisitionDimension = (
            "2D" if self.AcquisitionDimension == "MS" else self.AcquisitionDimension
        )
        self.AcqDateTime = str(
            datetime.strptime(hdr.general_info["exam_date"], "%Y.%m.%d / %H:%M:%S")
        )
        self.RepetitionTime = hdr.general_info["repetition_time"][0]
        self.ImageType = (
            ["ORIGINAL", "PRIMARY"]
            if hdr.general_info["recon_nr"] == 1
            else ["ORIGINAL", "SECONDARY"]
        )
        self.NumFiles = hdr.general_info["max_slices"]
        self.AcquisitionMatrix = [
            int(hdr.general_info["scan_resolution"][0]),
            int(hdr.general_info["scan_resolution"][1]),
        ]

        image_defs = hdr.image_defs.view(np.recarray)
        self.ImageOrientationPatient = ImageOrientation(
            np.concatenate([hdr.general_info["angulation"], [image_defs.slice_orientation[0]]])
        )
        self.ImagePositionPatient = TruncatedImageValue(hdr.general_info["off_center"])
        self.SliceOrientation = self.ImageOrientationPatient.get_plane()
        self.EchoTime = float(image_defs.echo_time[0])
        self.FlipAngle = float(image_defs.image_flip_angle[0])
        self.EchoTrainLength = (
            int(image_defs.turbo_factor[0]) / int(hdr.general_info["max_echoes"])
            if image_defs.turbo_factor[0] > 0
            else 1
        )
        self.InversionTime = float(image_defs.inversion_delay[0])
        self.ExContrastAgent = (
            image_defs.contrast_bolus_agent[0].decode("UTF-8")
            if "contrast_bolus_agent" in hdr.image_defs.dtype.fields.keys()
            else ""
        )
        self.ComplexImageComponent = COMPLEX_IMAGE_TYPES.get(
            image_defs.image_type_mr[0], "MAGNITUDE"
        )
        self.SliceThickness = float(image_defs.slice_thickness[0])
        self.SliceSpacing = float(image_defs.slice_thickness[0]) + float(image_defs.slice_gap[0])
        self.ReconMatrix = [
            int(image_defs.recon_resolution[0][0]),
            int(image_defs.recon_resolution[0][1]),
        ]
        self.ReconResolution = [
            float(image_defs.pixel_spacing[0][0]),
            float(image_defs.pixel_spacing[0][1]),
        ]
        self.FieldOfView = [res * num for res, num in zip(self.ReconResolution, self.ReconMatrix)]
        self.AcquiredResolution = [
            fov / num for fov, num in zip(self.FieldOfView, self.AcquisitionMatrix)
        ]
        self.NumberOfAverages = int(image_defs.number_of_averages[0])
        if manual_args is not None:
            for key, value in manual_args.items():
                setattr(self, key, value)

        if self.Truncated:
            logging.warning(
                "PAR header shows truncated information for image (%s)." % self.SourcePath
            )


class ParrecSet(BaseSet):
    def __init__(
        self,
        source: Path,
        output_root: Path,
        metadata_obj: Metadata,
        lut_obj: LookupTable,
        remove_identifiers: bool = False,
        date_shift_days: int = 0,
        manual_names: Optional[dict] = None,
        manual_args: Optional[dict] = None,
        input_hash: Optional[str] = None,
    ) -> None:
        super().__init__(
            source,
            output_root,
            metadata_obj,
            lut_obj,
            remove_identifiers,
            date_shift_days,
            manual_names,
            input_hash,
        )
        self.ManualArgs = manual_args

        logging.info("Loading PARRECs.")
        for parfile in sorted((output_root / self.Metadata.dir_to_str() / "parrec").rglob("*.par")):
            self.SeriesList.append(ParrecInfo(parfile, self.ManualArgs))

        study_nums, series_nums = self.get_unique_study_series(self.SeriesList)
        for di in self.SeriesList:
            logging.info("Processing %s" % di.SourcePath)
            if di.should_convert():
                if di.ReconstructionNumber > 1:
                    other_recons = [
                        other_di
                        for other_di in self.SeriesList
                        if other_di.SeriesNumber == di.SeriesNumber
                    ]
                    if any([other_di.ReconstructionNumber == 1 for other_di in other_recons]):
                        di.ConvertImage = False
                    elif not di.SeriesDescription.startswith("sWIP"):
                        di.ConvertImage = False
                if di.ConvertImage:
                    di.create_image_name(
                        self.Metadata.prefix_to_str(),
                        study_nums[di.SourcePath],
                        series_nums[di.SourcePath],
                        self.LookupTable,
                        self.ManualNames,
                    )

        logging.info("Generating unique names")
        self.generate_unique_names()


def sort_parrecs(parrec_dir: Path) -> None:
    logging.info("Sorting PARRECs")
    new_files = []
    study_uid = "2.25." + str(int(str(secrets.randbits(96))))
    pattern = re.compile(r"2\.25\.[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+\.par")
    for parfile in sorted(parrec_dir.rglob("*.par")):
        if pattern.search(parfile.name) is None:
            new_files.extend(split_fix_parrec(parfile, study_uid, parrec_dir))
            silentremove(parfile)
            silentremove(parfile.with_suffix(".rec"))
        else:
            new_files.append(parfile)
            new_files.append(parfile.with_suffix(".rec"))
    for path in parrec_dir.glob("*"):
        if path.name not in new_files:
            if path.is_dir():
                shutil.rmtree(path)
            else:
                path.unlink()
    logging.info("Sorting complete")



from pathlib import Path
from setuptools import setup, find_packages

__package_name__ = "radifox"


def get_version_and_cmdclass(pkg_path):
    """Load version.py module without importing the whole package.

    Template code from miniver
    """
    import os
    from importlib.util import module_from_spec, spec_from_file_location

    spec = spec_from_file_location("version", os.path.join(pkg_path, "_version.py"))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.get_cmdclass(pkg_path)


__version__, cmdclass = get_version_and_cmdclass(__package_name__)

setup(
    name=__package_name__,
    version=__version__,
    description=(
        "RADIFOX is the RADiological Image File Ontology eXtension, "
        "a Python package for the organization and management of medical images."
    ),
    long_description=(Path(__file__).parent.resolve() / "README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Blake Dewey",
    author_email="blake.dewey@jhu.edu",
    url="https://github.com/jh-mipc/radifox",
    license="Apache License, 2.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
    ],
    packages=find_packages(),
    keywords="mri conversion",
    entry_points={
        "console_scripts": [
            "radifox-convert=radifox.conversion.cli:convert",
            "radifox-update=radifox.conversion.cli:update",
            "radifox-qa=radifox.qa.run:run",
            "radifox-stage=radifox.modules.staging:Staging",
        ]
    },
    python_requires=">=3.10",
    install_requires=[
        "nibabel",
        "numpy",
        "pillow",
        "pydicom",
        "pyyaml",
        "scipy",
        "matplotlib",
        "trimesh",
        "networkx",
        "radifox-utils>=1.0.2",
    ],
    extras_require={
        "qa": [
            "flask",
            "gunicorn",
        ]
    },
    package_data={
        "radifox": [
            "conversion/parrec_templates/*.txt",
            "qa/templates/*.html",
        ]
    },
    cmdclass=cmdclass,
)
